"""
======
日志
======
AliceBot 使用 `loguru`_ 来记录日志信息。
自定义 logger 请参考 `loguru`_ 文档。

.. _loguru: https://github.com/Delgan/loguru
"""
from loguru import logger as _logger

logger = _logger
# logger = logging.getLogger("AliceBot")
